"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.worker = void 0;
const backfill_1 = require("../cache/backfill");
const getDepsForTarget_1 = require("../task/getDepsForTarget");
const taskId_1 = require("../task/taskId");
const getWorkspace_1 = require("../workspace/getWorkspace");
const WorkerQueue_1 = require("../task/WorkerQueue");
const logger_1 = require("../logger");
const Pipeline_1 = require("../task/Pipeline");
const TaskLogger_1 = require("../logger/TaskLogger");
const p_limit_1 = __importDefault(require("p-limit"));
/** Arbitrary parallel limit for network downloads of caches */
const NetworkParallelLimit = 15;
/** Worker command */
async function worker(cwd, config) {
    logger_1.logger.verbose("worker started");
    const workspace = (0, getWorkspace_1.getWorkspace)(cwd, config);
    const pipeline = new Pipeline_1.Pipeline(workspace, config);
    const networkLimit = (0, p_limit_1.default)(NetworkParallelLimit);
    /** the worker queue processor */
    const processor = async (job) => {
        const id = job.name;
        const { packageName, task } = (0, taskId_1.getPackageAndTask)(id);
        if (!pipeline.targets.has(id)) {
            logger_1.logger.info(`pipeline doesn't have  ${id}`);
            return;
        }
        const target = pipeline.targets.get(id);
        const deps = (0, getDepsForTarget_1.getDepsForTarget)(id, pipeline.dependencies);
        const taskLogger = new TaskLogger_1.TaskLogger(packageName || "[GLOBAL]", task);
        taskLogger.info(`[${job.id}] started`);
        // Step 1. Fetch all dependent cached outputs
        await Promise.all(deps
            .filter((d) => d !== Pipeline_1.START_TARGET_ID)
            .map((depTargetId) => {
            return networkLimit(() => getCache(pipeline.targets.get(depTargetId), workspace.root, config));
        }));
        taskLogger.info(`[${job.id}] restored all dependent cache`);
        // Step 2. Fetch this job cache (maybe from a previous run)
        const cacheResult = await getCache(target, workspace.root, config);
        if (cacheResult.cacheHit) {
            taskLogger.info(`[${job.id}] skipped`);
            return { hash: cacheResult.hash, id: target.id, cwd: target.cwd, outputGlob: target.outputGlob };
        }
        // Step 3. Try running the task based on the target ID from the queue
        try {
            let result;
            if (target.run) {
                if (target.packageName) {
                    result = target.run({
                        packageName: target.packageName,
                        config,
                        cwd: target.cwd,
                        options: target.options,
                        taskName: (0, taskId_1.getPackageAndTask)(target.id).task,
                        logger: taskLogger,
                    });
                }
                else {
                    result = target.run({
                        config,
                        cwd: target.cwd,
                        options: target.options,
                        logger: taskLogger,
                    });
                }
            }
            if (result && typeof result["then"] === "function") {
                await result;
            }
            // Step 4. Save the cache remotely (unconditionally)
            await saveCache(cacheResult.hash, target, config);
            taskLogger.info(`[${job.id}] done`);
            return { hash: cacheResult.hash, id: target.id, cwd: target.cwd, outputGlob: target.outputGlob };
        }
        catch (e) {
            // Step 4a. If there is an error, we report the output in a message as an Error. This gets sent
            //          across the redis server via the queue, so there will be a size limit to it
            taskLogger.error(`[${job.id}] failed`);
            throw new Error(taskLogger
                .getLogs()
                .map((e) => e.msg)
                .join("\n"));
        }
    };
    // After all the setup, we initialize the worker processor with the queue library
    const workerQueue = new WorkerQueue_1.WorkerQueue(config);
    await workerQueue.initializeAsWorker(processor);
    logger_1.logger.verbose("worker queue initialized");
}
exports.worker = worker;
/********************************************
 * Utilities to support caching of the worker
 *********************************************/
/**
 *
 * @param target
 * @param config
 * @returns
 */
function getCacheOptions(target, config) {
    return Object.assign(Object.assign({}, config.cacheOptions), { outputGlob: target.outputGlob || config.cacheOptions.outputGlob });
}
async function getCache(target, root, config) {
    let hash = null;
    let cacheHit = false;
    const { id, cwd } = target;
    const cacheOptions = getCacheOptions(target, config);
    hash = await (0, backfill_1.cacheHash)(id, root, cwd, cacheOptions, config.args);
    if (hash && !config.resetCache) {
        cacheHit = await (0, backfill_1.cacheFetch)(hash, id, cwd, config.cacheOptions);
    }
    return { hash, cacheHit };
}
async function saveCache(hash, target, config) {
    const cacheOptions = getCacheOptions(target, config);
    await (0, backfill_1.cachePut)(hash, target.cwd, cacheOptions);
}
